data = (
'jjwaels',    # 0x00
'jjwaelt',    # 0x01
'jjwaelp',    # 0x02
'jjwaelh',    # 0x03
'jjwaem',    # 0x04
'jjwaeb',    # 0x05
'jjwaebs',    # 0x06
'jjwaes',    # 0x07
'jjwaess',    # 0x08
'jjwaeng',    # 0x09
'jjwaej',    # 0x0a
'jjwaec',    # 0x0b
'jjwaek',    # 0x0c
'jjwaet',    # 0x0d
'jjwaep',    # 0x0e
'jjwaeh',    # 0x0f
'jjoe',    # 0x10
'jjoeg',    # 0x11
'jjoegg',    # 0x12
'jjoegs',    # 0x13
'jjoen',    # 0x14
'jjoenj',    # 0x15
'jjoenh',    # 0x16
'jjoed',    # 0x17
'jjoel',    # 0x18
'jjoelg',    # 0x19
'jjoelm',    # 0x1a
'jjoelb',    # 0x1b
'jjoels',    # 0x1c
'jjoelt',    # 0x1d
'jjoelp',    # 0x1e
'jjoelh',    # 0x1f
'jjoem',    # 0x20
'jjoeb',    # 0x21
'jjoebs',    # 0x22
'jjoes',    # 0x23
'jjoess',    # 0x24
'jjoeng',    # 0x25
'jjoej',    # 0x26
'jjoec',    # 0x27
'jjoek',    # 0x28
'jjoet',    # 0x29
'jjoep',    # 0x2a
'jjoeh',    # 0x2b
'jjyo',    # 0x2c
'jjyog',    # 0x2d
'jjyogg',    # 0x2e
'jjyogs',    # 0x2f
'jjyon',    # 0x30
'jjyonj',    # 0x31
'jjyonh',    # 0x32
'jjyod',    # 0x33
'jjyol',    # 0x34
'jjyolg',    # 0x35
'jjyolm',    # 0x36
'jjyolb',    # 0x37
'jjyols',    # 0x38
'jjyolt',    # 0x39
'jjyolp',    # 0x3a
'jjyolh',    # 0x3b
'jjyom',    # 0x3c
'jjyob',    # 0x3d
'jjyobs',    # 0x3e
'jjyos',    # 0x3f
'jjyoss',    # 0x40
'jjyong',    # 0x41
'jjyoj',    # 0x42
'jjyoc',    # 0x43
'jjyok',    # 0x44
'jjyot',    # 0x45
'jjyop',    # 0x46
'jjyoh',    # 0x47
'jju',    # 0x48
'jjug',    # 0x49
'jjugg',    # 0x4a
'jjugs',    # 0x4b
'jjun',    # 0x4c
'jjunj',    # 0x4d
'jjunh',    # 0x4e
'jjud',    # 0x4f
'jjul',    # 0x50
'jjulg',    # 0x51
'jjulm',    # 0x52
'jjulb',    # 0x53
'jjuls',    # 0x54
'jjult',    # 0x55
'jjulp',    # 0x56
'jjulh',    # 0x57
'jjum',    # 0x58
'jjub',    # 0x59
'jjubs',    # 0x5a
'jjus',    # 0x5b
'jjuss',    # 0x5c
'jjung',    # 0x5d
'jjuj',    # 0x5e
'jjuc',    # 0x5f
'jjuk',    # 0x60
'jjut',    # 0x61
'jjup',    # 0x62
'jjuh',    # 0x63
'jjweo',    # 0x64
'jjweog',    # 0x65
'jjweogg',    # 0x66
'jjweogs',    # 0x67
'jjweon',    # 0x68
'jjweonj',    # 0x69
'jjweonh',    # 0x6a
'jjweod',    # 0x6b
'jjweol',    # 0x6c
'jjweolg',    # 0x6d
'jjweolm',    # 0x6e
'jjweolb',    # 0x6f
'jjweols',    # 0x70
'jjweolt',    # 0x71
'jjweolp',    # 0x72
'jjweolh',    # 0x73
'jjweom',    # 0x74
'jjweob',    # 0x75
'jjweobs',    # 0x76
'jjweos',    # 0x77
'jjweoss',    # 0x78
'jjweong',    # 0x79
'jjweoj',    # 0x7a
'jjweoc',    # 0x7b
'jjweok',    # 0x7c
'jjweot',    # 0x7d
'jjweop',    # 0x7e
'jjweoh',    # 0x7f
'jjwe',    # 0x80
'jjweg',    # 0x81
'jjwegg',    # 0x82
'jjwegs',    # 0x83
'jjwen',    # 0x84
'jjwenj',    # 0x85
'jjwenh',    # 0x86
'jjwed',    # 0x87
'jjwel',    # 0x88
'jjwelg',    # 0x89
'jjwelm',    # 0x8a
'jjwelb',    # 0x8b
'jjwels',    # 0x8c
'jjwelt',    # 0x8d
'jjwelp',    # 0x8e
'jjwelh',    # 0x8f
'jjwem',    # 0x90
'jjweb',    # 0x91
'jjwebs',    # 0x92
'jjwes',    # 0x93
'jjwess',    # 0x94
'jjweng',    # 0x95
'jjwej',    # 0x96
'jjwec',    # 0x97
'jjwek',    # 0x98
'jjwet',    # 0x99
'jjwep',    # 0x9a
'jjweh',    # 0x9b
'jjwi',    # 0x9c
'jjwig',    # 0x9d
'jjwigg',    # 0x9e
'jjwigs',    # 0x9f
'jjwin',    # 0xa0
'jjwinj',    # 0xa1
'jjwinh',    # 0xa2
'jjwid',    # 0xa3
'jjwil',    # 0xa4
'jjwilg',    # 0xa5
'jjwilm',    # 0xa6
'jjwilb',    # 0xa7
'jjwils',    # 0xa8
'jjwilt',    # 0xa9
'jjwilp',    # 0xaa
'jjwilh',    # 0xab
'jjwim',    # 0xac
'jjwib',    # 0xad
'jjwibs',    # 0xae
'jjwis',    # 0xaf
'jjwiss',    # 0xb0
'jjwing',    # 0xb1
'jjwij',    # 0xb2
'jjwic',    # 0xb3
'jjwik',    # 0xb4
'jjwit',    # 0xb5
'jjwip',    # 0xb6
'jjwih',    # 0xb7
'jjyu',    # 0xb8
'jjyug',    # 0xb9
'jjyugg',    # 0xba
'jjyugs',    # 0xbb
'jjyun',    # 0xbc
'jjyunj',    # 0xbd
'jjyunh',    # 0xbe
'jjyud',    # 0xbf
'jjyul',    # 0xc0
'jjyulg',    # 0xc1
'jjyulm',    # 0xc2
'jjyulb',    # 0xc3
'jjyuls',    # 0xc4
'jjyult',    # 0xc5
'jjyulp',    # 0xc6
'jjyulh',    # 0xc7
'jjyum',    # 0xc8
'jjyub',    # 0xc9
'jjyubs',    # 0xca
'jjyus',    # 0xcb
'jjyuss',    # 0xcc
'jjyung',    # 0xcd
'jjyuj',    # 0xce
'jjyuc',    # 0xcf
'jjyuk',    # 0xd0
'jjyut',    # 0xd1
'jjyup',    # 0xd2
'jjyuh',    # 0xd3
'jjeu',    # 0xd4
'jjeug',    # 0xd5
'jjeugg',    # 0xd6
'jjeugs',    # 0xd7
'jjeun',    # 0xd8
'jjeunj',    # 0xd9
'jjeunh',    # 0xda
'jjeud',    # 0xdb
'jjeul',    # 0xdc
'jjeulg',    # 0xdd
'jjeulm',    # 0xde
'jjeulb',    # 0xdf
'jjeuls',    # 0xe0
'jjeult',    # 0xe1
'jjeulp',    # 0xe2
'jjeulh',    # 0xe3
'jjeum',    # 0xe4
'jjeub',    # 0xe5
'jjeubs',    # 0xe6
'jjeus',    # 0xe7
'jjeuss',    # 0xe8
'jjeung',    # 0xe9
'jjeuj',    # 0xea
'jjeuc',    # 0xeb
'jjeuk',    # 0xec
'jjeut',    # 0xed
'jjeup',    # 0xee
'jjeuh',    # 0xef
'jjyi',    # 0xf0
'jjyig',    # 0xf1
'jjyigg',    # 0xf2
'jjyigs',    # 0xf3
'jjyin',    # 0xf4
'jjyinj',    # 0xf5
'jjyinh',    # 0xf6
'jjyid',    # 0xf7
'jjyil',    # 0xf8
'jjyilg',    # 0xf9
'jjyilm',    # 0xfa
'jjyilb',    # 0xfb
'jjyils',    # 0xfc
'jjyilt',    # 0xfd
'jjyilp',    # 0xfe
'jjyilh',    # 0xff
)



# defusedxml
#
# Copyright (c) 2013 by Christian Heimes <christian@python.org>
# Licensed to PSF under a Contributor Agreement.
# See https://www.python.org/psf/license for licensing details.
"""Defused xml.etree.cElementTree
"""
from __future__ import absolute_import

from xml.etree.cElementTree import TreeBuilder as _TreeBuilder
from xml.etree.cElementTree import parse as _parse
from xml.etree.cElementTree import tostring

# iterparse from ElementTree!
from xml.etree.ElementTree import iterparse as _iterparse

from .ElementTree import DefusedXMLParser
from .common import _generate_etree_functions

__origin__ = "xml.etree.cElementTree"


# XMLParse is a typo, keep it for backwards compatibility
XMLTreeBuilder = XMLParse = XMLParser = DefusedXMLParser

parse, iterparse, fromstring = _generate_etree_functions(
    DefusedXMLParser, _TreeBuilder, _parse, _iterparse
)
XML = fromstring

__all__ = [
    "XML",
    "XMLParse",
    "XMLParser",
    "XMLTreeBuilder",
    "fromstring",
    "iterparse",
    "parse",
    "tostring",
]



# nif.py - functions for handling Spanish NIF (VAT) numbers
# coding: utf-8
#
# Copyright (C) 2012, 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIF (Número de Identificación Fiscal, Spanish VAT number).

The Spanish VAT number is a 9-digit number where either the first, last
digits or both can be letters.

The number is either a DNI (Documento Nacional de Identidad, for
Spaniards), a NIE (Número de Identificación de Extranjero, for
foreigners) or a CIF (Código de Identificación Fiscal, for legal
entities and others).

>>> compact('ES B-58378431')
'B58378431'
>>> validate('B64717838')
'B64717838'
>>> validate('B64717839')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('54362315K')  # resident
'54362315K'
>>> validate('X-5253868-R')  # foreign person
'X5253868R'
>>> validate('M-1234567-L')  # foreign person without NIE
'M1234567L'
"""

from stdnum.es import cif, dni, nie
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -').upper().strip()
    if number.startswith('ES'):
        number = number[2:]
    return number


def validate(number):
    """Check if the number provided is a valid VAT number. This checks the
    length, formatting and check digit."""
    number = compact(number)
    if not isdigits(number[1:-1]):
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    if number[0] in 'KLM':
        # K: Spanish younger than 14 year old
        # L: Spanish living outside Spain without DNI
        # M: granted the tax to foreigners who have no NIE
        # these use the old checkdigit algorithm (the DNI one)
        if number[-1] != dni.calc_check_digit(number[1:-1]):
            raise InvalidChecksum()
    elif isdigits(number[0]):
        # natural resident
        dni.validate(number)
    elif number[0] in 'XYZ':
        # foreign natural person
        nie.validate(number)
    else:
        # otherwise it has to be a valid CIF
        cif.validate(number)
    return number


def is_valid(number):
    """Check if the number provided is a valid VAT number. This checks the
    length, formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from typing import List  # noqa
from ._CustomOrderProperty import CustomOrderProperty  # noqa
import math


class Order(object):
    def __init__(self, order_id, total_gross, tip, pos_id, custom_order_properties, due_amount):
        # type: (str, float, float, str, List[CustomOrderProperty], float) -> None
        self.__order_id = order_id
        self.__total_gross = total_gross
        self.__tip = tip
        self.__pos_id = pos_id
        self.__custom_order_properties = custom_order_properties
        self.__due_amount = due_amount

    @property
    def order_id(self):
        # type: () -> str
        return self.__order_id

    @property
    def total_gross(self):
        # type: () -> float
        return self.__total_gross

    @property
    def tip(self):
        # type: () -> float
        return self.total_amount

    @property
    def pos_id(self):
        # type: () -> str
        return self.__pos_id

    @property
    def tip_percentage(self):
        # type: () -> float
        tip_percentage = (self.__tip * 100) / self.__total_gross
        return math.floor(tip_percentage * 100)/100.0

    @property
    def custom_order_properties(self):
        # type: () -> List[CustomOrderProperty]
        return self.__custom_order_properties

    def get_custom_property_by_key(self, key):
        custom_properties = filter(lambda c_o: c_o.key == key, self.custom_order_properties)
        if len(custom_properties) > 0:
            return custom_properties[0].value

        return None

    @property
    def due_amount(self):
        # type: () -> float
        return self.__due_amount



#
# The Python Imaging Library.
# $Id$
#
# EPS file handling
#
# History:
# 1995-09-01 fl   Created (0.1)
# 1996-05-18 fl   Don't choke on "atend" fields, Ghostscript interface (0.2)
# 1996-08-22 fl   Don't choke on floating point BoundingBox values
# 1996-08-23 fl   Handle files from Macintosh (0.3)
# 2001-02-17 fl   Use 're' instead of 'regex' (Python 2.1) (0.4)
# 2003-09-07 fl   Check gs.close status (from Federico Di Gregorio) (0.5)
# 2014-05-07 e    Handling of EPS with binary preview and fixed resolution
#                 resizing
#
# Copyright (c) 1997-2003 by Secret Labs AB.
# Copyright (c) 1995-2003 by Fredrik Lundh
#
# See the README file for information on usage and redistribution.
#
from __future__ import annotations

import io
import os
import re
import subprocess
import sys
import tempfile

from . import Image, ImageFile
from ._binary import i32le as i32
from ._deprecate import deprecate

# --------------------------------------------------------------------


split = re.compile(r"^%%([^:]*):[ \t]*(.*)[ \t]*$")
field = re.compile(r"^%[%!\w]([^:]*)[ \t]*$")

gs_binary = None
gs_windows_binary = None


def has_ghostscript():
    global gs_binary, gs_windows_binary
    if gs_binary is None:
        if sys.platform.startswith("win"):
            if gs_windows_binary is None:
                import shutil

                for binary in ("gswin32c", "gswin64c", "gs"):
                    if shutil.which(binary) is not None:
                        gs_windows_binary = binary
                        break
                else:
                    gs_windows_binary = False
            gs_binary = gs_windows_binary
        else:
            try:
                subprocess.check_call(["gs", "--version"], stdout=subprocess.DEVNULL)
                gs_binary = "gs"
            except OSError:
                gs_binary = False
    return gs_binary is not False


def Ghostscript(tile, size, fp, scale=1, transparency=False):
    """Render an image using Ghostscript"""
    global gs_binary
    if not has_ghostscript():
        msg = "Unable to locate Ghostscript on paths"
        raise OSError(msg)

    # Unpack decoder tile
    decoder, tile, offset, data = tile[0]
    length, bbox = data

    # Hack to support hi-res rendering
    scale = int(scale) or 1
    width = size[0] * scale
    height = size[1] * scale
    # resolution is dependent on bbox and size
    res_x = 72.0 * width / (bbox[2] - bbox[0])
    res_y = 72.0 * height / (bbox[3] - bbox[1])

    out_fd, outfile = tempfile.mkstemp()
    os.close(out_fd)

    infile_temp = None
    if hasattr(fp, "name") and os.path.exists(fp.name):
        infile = fp.name
    else:
        in_fd, infile_temp = tempfile.mkstemp()
        os.close(in_fd)
        infile = infile_temp

        # Ignore length and offset!
        # Ghostscript can read it
        # Copy whole file to read in Ghostscript
        with open(infile_temp, "wb") as f:
            # fetch length of fp
            fp.seek(0, io.SEEK_END)
            fsize = fp.tell()
            # ensure start position
            # go back
            fp.seek(0)
            lengthfile = fsize
            while lengthfile > 0:
                s = fp.read(min(lengthfile, 100 * 1024))
                if not s:
                    break
                lengthfile -= len(s)
                f.write(s)

    device = "pngalpha" if transparency else "ppmraw"

    # Build Ghostscript command
    command = [
        gs_binary,
        "-q",  # quiet mode
        f"-g{width:d}x{height:d}",  # set output geometry (pixels)
        f"-r{res_x:f}x{res_y:f}",  # set input DPI (dots per inch)
        "-dBATCH",  # exit after processing
        "-dNOPAUSE",  # don't pause between pages
        "-dSAFER",  # safe mode
        f"-sDEVICE={device}",
        f"-sOutputFile={outfile}",  # output file
        # adjust for image origin
        "-c",
        f"{-bbox[0]} {-bbox[1]} translate",
        "-f",
        infile,  # input file
        # showpage (see https://bugs.ghostscript.com/show_bug.cgi?id=698272)
        "-c",
        "showpage",
    ]

    # push data through Ghostscript
    try:
        startupinfo = None
        if sys.platform.startswith("win"):
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        subprocess.check_call(command, startupinfo=startupinfo)
        out_im = Image.open(outfile)
        out_im.load()
    finally:
        try:
            os.unlink(outfile)
            if infile_temp:
                os.unlink(infile_temp)
        except OSError:
            pass

    im = out_im.im.copy()
    out_im.close()
    return im


class PSFile:
    """
    Wrapper for bytesio object that treats either CR or LF as end of line.
    This class is no longer used internally, but kept for backwards compatibility.
    """

    def __init__(self, fp):
        deprecate(
            "PSFile",
            11,
            action="If you need the functionality of this class "
            "you will need to implement it yourself.",
        )
        self.fp = fp
        self.char = None

    def seek(self, offset, whence=io.SEEK_SET):
        self.char = None
        self.fp.seek(offset, whence)

    def readline(self):
        s = [self.char or b""]
        self.char = None

        c = self.fp.read(1)
        while (c not in b"\r\n") and len(c):
            s.append(c)
            c = self.fp.read(1)

        self.char = self.fp.read(1)
        # line endings can be 1 or 2 of \r \n, in either order
        if self.char in b"\r\n":
            self.char = None

        return b"".join(s).decode("latin-1")


def _accept(prefix):
    return prefix[:4] == b"%!PS" or (len(prefix) >= 4 and i32(prefix) == 0xC6D3D0C5)


##
# Image plugin for Encapsulated PostScript. This plugin supports only
# a few variants of this format.


class EpsImageFile(ImageFile.ImageFile):
    """EPS File Parser for the Python Imaging Library"""

    format = "EPS"
    format_description = "Encapsulated Postscript"

    mode_map = {1: "L", 2: "LAB", 3: "RGB", 4: "CMYK"}

    def _open(self):
        (length, offset) = self._find_offset(self.fp)

        # go to offset - start of "%!PS"
        self.fp.seek(offset)

        self._mode = "RGB"
        self._size = None

        byte_arr = bytearray(255)
        bytes_mv = memoryview(byte_arr)
        bytes_read = 0
        reading_header_comments = True
        reading_trailer_comments = False
        trailer_reached = False

        def check_required_header_comments():
            if "PS-Adobe" not in self.info:
                msg = 'EPS header missing "%!PS-Adobe" comment'
                raise SyntaxError(msg)
            if "BoundingBox" not in self.info:
                msg = 'EPS header missing "%%BoundingBox" comment'
                raise SyntaxError(msg)

        def _read_comment(s):
            nonlocal reading_trailer_comments
            try:
                m = split.match(s)
            except re.error as e:
                msg = "not an EPS file"
                raise SyntaxError(msg) from e

            if m:
                k, v = m.group(1, 2)
                self.info[k] = v
                if k == "BoundingBox":
                    if v == "(atend)":
                        reading_trailer_comments = True
                    elif not self._size or (
                        trailer_reached and reading_trailer_comments
                    ):
                        try:
                            # Note: The DSC spec says that BoundingBox
                            # fields should be integers, but some drivers
                            # put floating point values there anyway.
                            box = [int(float(i)) for i in v.split()]
                            self._size = box[2] - box[0], box[3] - box[1]
                            self.tile = [
                                ("eps", (0, 0) + self.size, offset, (length, box))
                            ]
                        except Exception:
                            pass
                return True

        while True:
            byte = self.fp.read(1)
            if byte == b"":
                # if we didn't read a byte we must be at the end of the file
                if bytes_read == 0:
                    break
            elif byte in b"\r\n":
                # if we read a line ending character, ignore it and parse what
                # we have already read. if we haven't read any other characters,
                # continue reading
                if bytes_read == 0:
                    continue
            else:
                # ASCII/hexadecimal lines in an EPS file must not exceed
                # 255 characters, not including line ending characters
                if bytes_read >= 255:
                    # only enforce this for lines starting with a "%",
                    # otherwise assume it's binary data
                    if byte_arr[0] == ord("%"):
                        msg = "not an EPS file"
                        raise SyntaxError(msg)
                    else:
                        if reading_header_comments:
                            check_required_header_comments()
                            reading_header_comments = False
                        # reset bytes_read so we can keep reading
                        # data until the end of the line
                        bytes_read = 0
                byte_arr[bytes_read] = byte[0]
                bytes_read += 1
                continue

            if reading_header_comments:
                # Load EPS header

                # if this line doesn't start with a "%",
                # or does start with "%%EndComments",
                # then we've reached the end of the header/comments
                if byte_arr[0] != ord("%") or bytes_mv[:13] == b"%%EndComments":
                    check_required_header_comments()
                    reading_header_comments = False
                    continue

                s = str(bytes_mv[:bytes_read], "latin-1")
                if not _read_comment(s):
                    m = field.match(s)
                    if m:
                        k = m.group(1)
                        if k[:8] == "PS-Adobe":
                            self.info["PS-Adobe"] = k[9:]
                        else:
                            self.info[k] = ""
                    elif s[0] == "%":
                        # handle non-DSC PostScript comments that some
                        # tools mistakenly put in the Comments section
                        pass
                    else:
                        msg = "bad EPS header"
                        raise OSError(msg)
            elif bytes_mv[:11] == b"%ImageData:":
                # Check for an "ImageData" descriptor
                # https://www.adobe.com/devnet-apps/photoshop/fileformatashtml/#50577413_pgfId-1035096

                # Values:
                # columns
                # rows
                # bit depth (1 or 8)
                # mode (1: L, 2: LAB, 3: RGB, 4: CMYK)
                # number of padding channels
                # block size (number of bytes per row per channel)
                # binary/ascii (1: binary, 2: ascii)
                # data start identifier (the image data follows after a single line
                #   consisting only of this quoted value)
                image_data_values = byte_arr[11:bytes_read].split(None, 7)
                columns, rows, bit_depth, mode_id = (
                    int(value) for value in image_data_values[:4]
                )

                if bit_depth == 1:
                    self._mode = "1"
                elif bit_depth == 8:
                    try:
                        self._mode = self.mode_map[mode_id]
                    except ValueError:
                        break
                else:
                    break

                self._size = columns, rows
                return
            elif trailer_reached and reading_trailer_comments:
                # Load EPS trailer

                # if this line starts with "%%EOF",
                # then we've reached the end of the file
                if bytes_mv[:5] == b"%%EOF":
                    break

                s = str(bytes_mv[:bytes_read], "latin-1")
                _read_comment(s)
            elif bytes_mv[:9] == b"%%Trailer":
                trailer_reached = True
            bytes_read = 0

        check_required_header_comments()

        if not self._size:
            msg = "cannot determine EPS bounding box"
            raise OSError(msg)

    def _find_offset(self, fp):
        s = fp.read(4)

        if s == b"%!PS":
            # for HEAD without binary preview
            fp.seek(0, io.SEEK_END)
            length = fp.tell()
            offset = 0
        elif i32(s) == 0xC6D3D0C5:
            # FIX for: Some EPS file not handled correctly / issue #302
            # EPS can contain binary data
            # or start directly with latin coding
            # more info see:
            # https://web.archive.org/web/20160528181353/http://partners.adobe.com/public/developer/en/ps/5002.EPSF_Spec.pdf
            s = fp.read(8)
            offset = i32(s)
            length = i32(s, 4)
        else:
            msg = "not an EPS file"
            raise SyntaxError(msg)

        return length, offset

    def load(self, scale=1, transparency=False):
        # Load EPS via Ghostscript
        if self.tile:
            self.im = Ghostscript(self.tile, self.size, self.fp, scale, transparency)
            self._mode = self.im.mode
            self._size = self.im.size
            self.tile = []
        return Image.Image.load(self)

    def load_seek(self, *args, **kwargs):
        # we can't incrementally load, so force ImageFile.parser to
        # use our custom load method by defining this method.
        pass


# --------------------------------------------------------------------


def _save(im, fp, filename, eps=1):
    """EPS Writer for the Python Imaging Library."""

    # make sure image data is available
    im.load()

    # determine PostScript image mode
    if im.mode == "L":
        operator = (8, 1, b"image")
    elif im.mode == "RGB":
        operator = (8, 3, b"false 3 colorimage")
    elif im.mode == "CMYK":
        operator = (8, 4, b"false 4 colorimage")
    else:
        msg = "image mode is not supported"
        raise ValueError(msg)

    if eps:
        # write EPS header
        fp.write(b"%!PS-Adobe-3.0 EPSF-3.0\n")
        fp.write(b"%%Creator: PIL 0.1 EpsEncode\n")
        # fp.write("%%CreationDate: %s"...)
        fp.write(b"%%%%BoundingBox: 0 0 %d %d\n" % im.size)
        fp.write(b"%%Pages: 1\n")
        fp.write(b"%%EndComments\n")
        fp.write(b"%%Page: 1 1\n")
        fp.write(b"%%ImageData: %d %d " % im.size)
        fp.write(b'%d %d 0 1 1 "%s"\n' % operator)

    # image header
    fp.write(b"gsave\n")
    fp.write(b"10 dict begin\n")
    fp.write(b"/buf %d string def\n" % (im.size[0] * operator[1]))
    fp.write(b"%d %d scale\n" % im.size)
    fp.write(b"%d %d 8\n" % im.size)  # <= bits
    fp.write(b"[%d 0 0 -%d 0 %d]\n" % (im.size[0], im.size[1], im.size[1]))
    fp.write(b"{ currentfile buf readhexstring pop } bind\n")
    fp.write(operator[2] + b"\n")
    if hasattr(fp, "flush"):
        fp.flush()

    ImageFile._save(im, fp, [("eps", (0, 0) + im.size, 0, None)])

    fp.write(b"\n%%%%EndBinary\n")
    fp.write(b"grestore end\n")
    if hasattr(fp, "flush"):
        fp.flush()


# --------------------------------------------------------------------


Image.register_open(EpsImageFile.format, EpsImageFile, _accept)

Image.register_save(EpsImageFile.format, _save)

Image.register_extensions(EpsImageFile.format, [".ps", ".eps"])

Image.register_mime(EpsImageFile.format, "application/postscript")



"""distutils.command.install_data

Implements the Distutils 'install_data' command, for installing
platform-independent data files."""

# contributed by Bastian Kleineidam

import os

from ..core import Command
from ..util import change_root, convert_path


class install_data(Command):
    description = "install data files"

    user_options = [
        (
            'install-dir=',
            'd',
            "base directory for installing data files "
            "(default: installation base dir)",
        ),
        ('root=', None, "install everything relative to this alternate root directory"),
        ('force', 'f', "force installation (overwrite existing files)"),
    ]

    boolean_options = ['force']

    def initialize_options(self):
        self.install_dir = None
        self.outfiles = []
        self.root = None
        self.force = 0
        self.data_files = self.distribution.data_files
        self.warn_dir = 1

    def finalize_options(self):
        self.set_undefined_options(
            'install',
            ('install_data', 'install_dir'),
            ('root', 'root'),
            ('force', 'force'),
        )

    def run(self):
        self.mkpath(self.install_dir)
        for f in self.data_files:
            if isinstance(f, str):
                # it's a simple file, so copy it
                f = convert_path(f)
                if self.warn_dir:
                    self.warn(
                        "setup script did not provide a directory for "
                        f"'{f}' -- installing right in '{self.install_dir}'"
                    )
                (out, _) = self.copy_file(f, self.install_dir)
                self.outfiles.append(out)
            else:
                # it's a tuple with path to install to and a list of files
                dir = convert_path(f[0])
                if not os.path.isabs(dir):
                    dir = os.path.join(self.install_dir, dir)
                elif self.root:
                    dir = change_root(self.root, dir)
                self.mkpath(dir)

                if f[1] == []:
                    # If there are no files listed, the user must be
                    # trying to create an empty directory, so add the
                    # directory to the list of output files.
                    self.outfiles.append(dir)
                else:
                    # Copy files, adding them to the list of output files.
                    for data in f[1]:
                        data = convert_path(data)
                        (out, _) = self.copy_file(data, dir)
                        self.outfiles.append(out)

    def get_inputs(self):
        return self.data_files or []

    def get_outputs(self):
        return self.outfiles



from actions import mb_context
from bustoken import create_token
from msgbus import TK_SYS_ACK
from typing import Optional

IS_DELIVERY_FEE_ENABLED = None  # type: Optional[bool]
TK_REMOTE_ORDER_CHECK_IF_DELIVERY_FEE_IS_ENABLED = create_token("0", "37", "224")


def _get_delivery_fee_enabled_status():
    # type: () -> None

    global IS_DELIVERY_FEE_ENABLED

    msg = mb_context.MB_EasySendMessage("RemoteOrder", TK_REMOTE_ORDER_CHECK_IF_DELIVERY_FEE_IS_ENABLED)
    if msg.token == TK_SYS_ACK:
        IS_DELIVERY_FEE_ENABLED = True if msg.data == "True" else False


def is_delivery_fee_enabled():
    # type: () -> bool

    if IS_DELIVERY_FEE_ENABLED is None:
        _get_delivery_fee_enabled_status()

    return IS_DELIVERY_FEE_ENABLED



from xml.etree import cElementTree as eTree

from messagebus import (
    DefaultToken,
    Message,
    MessageBus,
    TokenCreator,
    TokenPriority,
)
from nreports.model.exceptions import UserDataNotFound

TK_USERCTRL_GETINFO = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="12",
    number="2",
)  # 18874370


class UserRepository(object):

    def __init__(
        self,
        message_bus: MessageBus,
    ) -> None:

        self._message_bus = message_bus

    def get_user_information(
        self,
        user_id: str,
    ) -> eTree.Element:

        msg = self._message_bus.send_message(
            component="UserControl",
            message=Message(
                token=TK_USERCTRL_GETINFO,
                data=user_id.encode("utf-8"),
            ),
        )
        if msg.token == DefaultToken.TK_SYS_ACK:
            xml_response = eTree.XML(msg.data)
            return xml_response.find("user")

        raise RuntimeError("Could not retrieve info from the UserControl service")

    def get_user_name(
        self,
        user_id: str,
    ) -> str:

        user_data = self.get_user_information(user_id=user_id)
        if user_data is not None:
            return user_data.get("LongName")

        raise UserDataNotFound(user_id=user_id)



import os

from configurator import Configurator
from mbcontextmessagehandler import MbContextMessageBus, MbContextMessageHandler
from messagehandlerbuilder import ServiceMessageHandlerBuilder
from model.event import all_events, ListenedEvents
from msgbus import MBEasyContext

from helper import import_pydevd


def main():
    # type: () -> None

    loader_path = os.environ["LOADERCFG"]
    import_pydevd(loader_path, 9176, False)

    configurator = Configurator(loader_path)
    configs = configurator.read_loader_configs()

    logger = configurator.configure_logger()
    logger.setLevel(configs.log_level)

    mb_context = MBEasyContext(configs.component_name)
    message_bus = MbContextMessageBus(mb_context)
    service_message_handler_builder = ServiceMessageHandlerBuilder(logger, configs, message_bus)

    message_handler = MbContextMessageHandler(
        message_bus,
        configs.component_name,
        configs.component_name,
        "",
        service_message_handler_builder
    )

    events = all_events()
    events.remove(ListenedEvents.ORDER_MODIFIED)

    message_handler.subscribe_reentrant_events(events)

    logger.debug("Starting Message Bus [{}] ...".format(configs.component_name))
    return message_handler.handle_events()


if __name__ == "__main__":
    main()



from actions.delivery.delivery_address import (
    request_and_save_customer_data,
    request_customer_data,
)
from actions.delivery.get_delivery_fee_enabled_status import is_delivery_fee_enabled
from get_store_status import get_remote_order_status
from delivery_manual_order import create_delivery_manual_order_json
from produce_delivery_order import send_remote_order_to_production



from cfgtools import read


class ConfiguratorReader:
    def __init__(self, loader_cfg, service_name, manager_name):
        self.service_name = service_name
        self.manager_name = manager_name

        self.reader = read(loader_cfg)

    def get_configuration(
        self,                  # type: ConfiguratorReader
        param,                 # type: str
        default_value="",      # type: str
    ):
        # type: (...) -> str

        value = self.reader.find_value(self.service_name + "." + self.manager_name + "." + param)
        if value is None:
            return default_value

        return value

    def get_configuration_list(self, param):
        return self.reader.find_values(self.service_name + "." + self.manager_name + "." + param)



from pos_model import TaxItem, SaleItem


class TaxItemFormatter(object):
    pattern = "<TaxItem lineNumber=\"{line_number}\" " \
              "level=\"{level}\" " \
              "itemId=\"{item_id}\" " \
              "partCode=\"{part_code}\" " \
              "taxRuleId=\"{tax_rule_id}\" " \
              "taxName=\"{tax_name}\" " \
              "taxIndex=\"{tax_index}\" " \
              "baseAmountBD=\"{base_ammount_bd}\" " \
              "taxAmountBD=\"{tax_ammount_bd}\" " \
              "baseAmountAD=\"{base_ammount_ad}\" " \
              "taxAmountAD=\"{tax_ammount_ad}\" " \
              "taxIncluded=\"{tax_included}\" " \
              "taxRate=\"{tax_rate}\"/>"

    def __init__(self):
        return

    def format_tax_item(self, tax_item, sale_item):
        # type: (TaxItem, SaleItem) -> str
        if tax_item is None:
            raise ValueError("tax_item cannot be None")

        return self.pattern.format(line_number=sale_item.line_number,
                                   level=sale_item.level,
                                   item_id=sale_item.item_id,
                                   part_code=sale_item.part_code,
                                   tax_rule_id=tax_item.tax_rule_id,
                                   tax_name=tax_item.tax_name,
                                   tax_index=tax_item.tax_name + "|" + tax_item.tax_index,
                                   base_ammount_bd=tax_item.base_amount_bd,
                                   tax_ammount_bd=tax_item.tax_amount_bd,
                                   base_ammount_ad=tax_item.base_amount_ad,
                                   tax_ammount_ad=tax_item.tax_amount_ad,
                                   tax_included=tax_item.tax_included,
                                   tax_rate=tax_item.tax_rate)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    Sequence,
    TypeVar,
    overload,
)

from opentelemetry.util.types import AttributesAsKey, AttributeValue

_T = TypeVar("_T")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")

def ns_to_iso_str(nanoseconds: int) -> str: ...
def get_dict_as_key(
    labels: Mapping[str, AttributeValue]
) -> AttributesAsKey: ...

class BoundedList(Sequence[_T]):
    """An append only list with a fixed max size.

    Calls to `append` and `extend` will drop the oldest elements if there is
    not enough room.
    """

    dropped: int
    def __init__(self, maxlen: int): ...
    def insert(self, index: int, value: _T) -> None: ...
    @overload
    def __getitem__(self, i: int) -> _T: ...
    @overload
    def __getitem__(self, s: slice) -> Sequence[_T]: ...
    def __len__(self) -> int: ...
    def append(self, item: _T): ...
    def extend(self, seq: Sequence[_T]): ...
    @classmethod
    def from_seq(cls, maxlen: int, seq: Iterable[_T]) -> BoundedList[_T]: ...

class BoundedDict(MutableMapping[_KT, _VT]):
    """An ordered dict with a fixed max capacity.

    Oldest elements are dropped when the dict is full and a new element is
    added.
    """

    dropped: int
    def __init__(self, maxlen: int): ...
    def __getitem__(self, k: _KT) -> _VT: ...
    def __setitem__(self, k: _KT, v: _VT) -> None: ...
    def __delitem__(self, v: _KT) -> None: ...
    def __iter__(self) -> Iterator[_KT]: ...
    def __len__(self) -> int: ...
    @classmethod
    def from_map(
        cls, maxlen: int, mapping: Mapping[_KT, _VT]
    ) -> BoundedDict[_KT, _VT]: ...



"""Timestamp comparison of files and groups of files."""

import functools
import os.path

from ._functools import splat
from .errors import DistutilsFileError
from .py39compat import zip_strict


def _newer(source, target):
    return not os.path.exists(target) or (
        os.path.getmtime(source) > os.path.getmtime(target)
    )


def newer(source, target):
    """
    Is source modified more recently than target.

    Returns True if 'source' is modified more recently than
    'target' or if 'target' does not exist.

    Raises DistutilsFileError if 'source' does not exist.
    """
    if not os.path.exists(source):
        raise DistutilsFileError("file '%s' does not exist" % os.path.abspath(source))

    return _newer(source, target)


def newer_pairwise(sources, targets, newer=newer):
    """
    Filter filenames where sources are newer than targets.

    Walk two filename iterables in parallel, testing if each source is newer
    than its corresponding target.  Returns a pair of lists (sources,
    targets) where source is newer than target, according to the semantics
    of 'newer()'.
    """
    newer_pairs = filter(splat(newer), zip_strict(sources, targets))
    return tuple(map(list, zip(*newer_pairs))) or ([], [])


def newer_group(sources, target, missing='error'):
    """
    Is target out-of-date with respect to any file in sources.

    Return True if 'target' is out-of-date with respect to any file
    listed in 'sources'. In other words, if 'target' exists and is newer
    than every file in 'sources', return False; otherwise return True.
    ``missing`` controls how to handle a missing source file:

    - error (default): allow the ``stat()`` call to fail.
    - ignore: silently disregard any missing source files.
    - newer: treat missing source files as "target out of date". This
      mode is handy in "dry-run" mode: it will pretend to carry out
      commands that wouldn't work because inputs are missing, but
      that doesn't matter because dry-run won't run the commands.
    """

    def missing_as_newer(source):
        return missing == 'newer' and not os.path.exists(source)

    ignored = os.path.exists if missing == 'ignore' else None
    return any(
        missing_as_newer(source) or _newer(source, target)
        for source in filter(ignored, sources)
    )


newer_pairwise_group = functools.partial(newer_pairwise, newer=newer_group)



from abc import ABCMeta, abstractmethod
from logging import Logger
from threading import Thread

from production.model import ProductionOrder
from typing import List


class ProductionView(object):
    __metaclass__ = ABCMeta

    def __init__(self, name, logger):
        # type: (str, Logger) -> None
        self.name = name
        self.logger = logger
        if self.logger is not None:
            self.log_level = {
                "debug": self.logger.debug,
                "warning": self.logger.warning,
                "error": self.logger.error,
                "exception": self.logger.exception
            }

    @abstractmethod
    def handle_order(self, order, synch=False):
        # type: (ProductionOrder, bool) -> Thread

        raise NotImplementedError()

    @abstractmethod
    def get_view_tags(self):
        # type: () -> List[str]

        raise NotImplementedError()

    def debug(self, msg, *args):
        self.log("debug", msg, *args)

    def warning(self, msg, *args):
        self.log("warning", msg, *args)

    def error(self, msg, *args):
        self.log("error", msg, *args)

    def exception(self, msg, *args):
        self.log("exception", msg, *args)

    def log(self, level, msg, *args):
        if self.logger is not None:
            func = self.log_level[level]
            msg = msg.format(*args)
            func(msg)



class VariableNotGeneratedByBuilder(Exception):

    def __init__(
        self,
        variable_name: str,
    ) -> None:

        self.variable_name = variable_name

    def __str__(
        self
    ) -> str:

        return f"Error variable_name: {self.variable_name}"



from datetime import datetime
from typing import (
    Dict,
    Optional,
    Tuple,
)
from xml.etree import cElementTree as eTree
from mwhelper import ModelService
from typing import List
from externals.repository import AlertsRepository
from services.model import ConfiguredAlert
from .table import get_table_picture
from posentitiesformatter import (
    OrderFormatter,
    TableFormatter,
)


def get_configured_alerts_by_type(
    notify_type,        # type: str
    configured_alerts,  # type: list
):
    # type: (...) -> List[ConfiguredAlert]

    alerts = []
    for config in configured_alerts:
        if "notifyOn" in config.alert_json and notify_type == config.alert_json["notifyOn"]:
            alerts.append(config)

    return alerts


def insert_event_into_repository(
    pos_id,                   # type: str
    alerts_repository,        # type: AlertsRepository
    order_picture,            # type: eTree.XML
    alert,                    # type: ConfiguredAlert
    table_picture_dict=None,  # type: Optional[dict]
    order_picture_dict=None,  # type: Optional[dict]
):
    # type: (...) -> Tuple[Dict, Dict]

    timestamp = str(datetime.utcnow())
    notification_json = {
        "alertJSON": alert.alert_json,
        "timestamp": timestamp,
    }
    is_table = ModelService.is_table_service(order_picture)
    if order_picture_dict is None:
        if is_table and table_picture_dict is None:
            table_id = order_picture.find("CustomOrderProperties/OrderProperty[@key='TABLE_ID']").get("value")
            table_picture = eTree.XML(
                get_table_picture(
                    pos_id=pos_id,
                    table_id=table_id,
                )
            )
            table_picture_dict = TableFormatter().get_table_picture_dict(table_picture=table_picture)
            order_picture_dict = OrderFormatter().get_order_picture_dict(order_picture=order_picture)
        else:
            order_picture_dict = OrderFormatter().get_order_picture_dict(order_picture=order_picture)

    notification_json.update({"orderPicture": order_picture_dict})
    if is_table:
        notification_json.update({"tablePicture": table_picture_dict})

    alerts_repository.insert_event(
        alert_id=alert.id,
        notification_json=notification_json,
        timestamp=timestamp,
        sent_status=0,
    )

    return order_picture_dict, table_picture_dict



import unittest
import importlib_resources as resources

from . import data01
from . import util


class ContentsTests:
    expected = {
        '__init__.py',
        'binary.file',
        'subdirectory',
        'utf-16.file',
        'utf-8.file',
    }

    def test_contents(self):
        contents = {path.name for path in resources.files(self.data).iterdir()}
        assert self.expected <= contents


class ContentsDiskTests(ContentsTests, unittest.TestCase):
    def setUp(self):
        self.data = data01


class ContentsZipTests(ContentsTests, util.ZipSetup, unittest.TestCase):
    pass


class ContentsNamespaceTests(ContentsTests, unittest.TestCase):
    expected = {
        # no __init__ because of namespace design
        # no subdirectory as incidental difference in fixture
        'binary.file',
        'utf-16.file',
        'utf-8.file',
    }

    def setUp(self):
        from . import namespacedata01

        self.data = namespacedata01



# handelsregisternummer.py - functions for handling German company registry id
# coding: utf-8
#
# Copyright (C) 2015 Holvi Payment Services Oy
# Copyright (C) 2018-2022 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Handelsregisternummer (German company register number).

The number consists of the court where the company has registered, the type
of register and the registration number.

The type of the register is either HRA or HRB where the letter "B" stands for
HR section B, where limited liability companies and corporations are entered
(GmbH's and AG's). There is also a section HRA for business partnerships
(OHG's, KG's etc.). In other words: businesses in section HRB are limited
liability companies, while businesses in HRA have personally liable partners.

More information:

* https://www.handelsregister.de/
* https://en.wikipedia.org/wiki/German_Trade_Register
* https://offeneregister.de/

>>> validate('Aachen HRA 11223')
'Aachen HRA 11223'
>>> validate('Frankfurt/Oder GnR 11223', company_form='e.G.')
'Frankfurt/Oder GnR 11223'
>>> validate('Aachen HRC 44123')
Traceback (most recent call last):
  ...
InvalidFormat: ...
>>> validate('Aachen HRA 44123', company_form='GmbH')
Traceback (most recent call last):
  ...
InvalidComponent: ...
"""

import re
import unicodedata

from stdnum.exceptions import *
from stdnum.util import clean, to_unicode


# The known courts that have a Handelsregister
GERMAN_COURTS = (
    'Aachen',
    'Altenburg',
    'Amberg',
    'Ansbach',
    'Apolda',
    'Arnsberg',
    'Arnstadt Zweigstelle Ilmenau',
    'Arnstadt',
    'Aschaffenburg',
    'Augsburg',
    'Aurich',
    'Bad Hersfeld',
    'Bad Homburg v.d.H.',
    'Bad Kreuznach',
    'Bad Oeynhausen',
    'Bad Salzungen',
    'Bamberg',
    'Bayreuth',
    'Berlin (Charlottenburg)',
    'Bielefeld',
    'Bochum',
    'Bonn',
    'Braunschweig',
    'Bremen',
    'Chemnitz',
    'Coburg',
    'Coesfeld',
    'Cottbus',
    'Darmstadt',
    'Deggendorf',
    'Dortmund',
    'Dresden',
    'Duisburg',
    'Düren',
    'Düsseldorf',
    'Eisenach',
    'Erfurt',
    'Eschwege',
    'Essen',
    'Flensburg',
    'Frankfurt am Main',
    'Frankfurt/Oder',
    'Freiburg',
    'Friedberg',
    'Fritzlar',
    'Fulda',
    'Fürth',
    'Gelsenkirchen',
    'Gera',
    'Gießen',
    'Gotha',
    'Greiz',
    'Göttingen',
    'Gütersloh',
    'Hagen',
    'Hamburg',
    'Hamm',
    'Hanau',
    'Hannover',
    'Heilbad Heiligenstadt',
    'Hildburghausen',
    'Hildesheim',
    'Hof',
    'Homburg',
    'Ingolstadt',
    'Iserlohn',
    'Jena',
    'Kaiserslautern',
    'Kassel',
    'Kempten (Allgäu)',
    'Kiel',
    'Kleve',
    'Koblenz',
    'Korbach',
    'Krefeld',
    'Köln',
    'Königstein',
    'Landau',
    'Landshut',
    'Langenfeld',
    'Lebach',
    'Leipzig',
    'Lemgo',
    'Limburg',
    'Ludwigshafen a.Rhein (Ludwigshafen)',
    'Lübeck',
    'Lüneburg',
    'Mainz',
    'Mannheim',
    'Marburg',
    'Meiningen',
    'Memmingen',
    'Merzig',
    'Montabaur',
    'Mönchengladbach',
    'Mühlhausen',
    'München',
    'Münster',
    'Neubrandenburg',
    'Neunkirchen',
    'Neuruppin',
    'Neuss',
    'Nordhausen',
    'Nürnberg',
    'Offenbach am Main',
    'Oldenburg (Oldenburg)',
    'Osnabrück',
    'Ottweiler',
    'Paderborn',
    'Passau',
    'Pinneberg',
    'Potsdam',
    'Pößneck Zweigstelle Bad Lobenstein',
    'Pößneck',
    'Recklinghausen',
    'Regensburg',
    'Rostock',
    'Rudolstadt Zweigstelle Saalfeld',
    'Rudolstadt',
    'Saarbrücken',
    'Saarlouis',
    'Schweinfurt',
    'Schwerin',
    'Siegburg',
    'Siegen',
    'Sondershausen',
    'Sonneberg',
    'St. Ingbert (St Ingbert)',
    'St. Wendel (St Wendel)',
    'Stadthagen',
    'Stadtroda',
    'Steinfurt',
    'Stendal',
    'Stralsund',
    'Straubing',
    'Stuttgart',
    'Suhl',
    'Sömmerda',
    'Tostedt',
    'Traunstein',
    'Ulm',
    'Völklingen',
    'Walsrode',
    'Weiden i. d. OPf.',
    'Weimar',
    'Wetzlar',
    'Wiesbaden',
    'Wittlich',
    'Wuppertal',
    'Würzburg',
    'Zweibrücken',
)


def _to_min(court):
    """Convert the court name for quick comparison without encoding issues."""
    return ''.join(
        x for x in unicodedata.normalize('NFD', to_unicode(court).lower())
        if x in 'abcdefghijklmnopqrstuvwxyz')


# Build a dictionary for lookup up courts
_courts = dict(
    (_to_min(court), court) for court in GERMAN_COURTS)
_courts.update(
    (_to_min(alias), court) for alias, court in (
        ('Allgäu', 'Kempten (Allgäu)'),
        ('Bad Homburg', 'Bad Homburg v.d.H.'),
        ('Berlin', 'Berlin (Charlottenburg)'),
        ('Charlottenburg', 'Berlin (Charlottenburg)'),
        ('Charlottenburg (Berlin)', 'Berlin (Charlottenburg)'),
        ('Kaln', 'Köln'),  # for encoding issues
        ('Kempten', 'Kempten (Allgäu)'),
        ('Ludwigshafen am Rhein (Ludwigshafen)', 'Ludwigshafen a.Rhein (Ludwigshafen)'),
        ('Ludwigshafen am Rhein', 'Ludwigshafen a.Rhein (Ludwigshafen)'),
        ('Ludwigshafen', 'Ludwigshafen a.Rhein (Ludwigshafen)'),
        ('Oldenburg', 'Oldenburg (Oldenburg)'),
        ('St. Ingbert', 'St. Ingbert (St Ingbert)'),
        ('St. Wendel', 'St. Wendel (St Wendel)'),
        ('Weiden in der Oberpfalz', 'Weiden i. d. OPf.'),
        ('Weiden', 'Weiden i. d. OPf.'),
        ('Paderborn früher Höxter', 'Paderborn'),
    ))


# The known registry types
REGISTRY_TYPES = (
    'HRA',
    'HRB',
    'PR',
    'GnR',
    'VR',
)

COMPANY_FORM_REGISTRY_TYPES = {
    'e.K.': 'HRA',
    'e.V.': 'VR',
    'Verein': 'VR',
    'OHG': 'HRA',
    'KG': 'HRA',
    'KGaA': 'HRB',
    'Vor-GmbH': 'HRB',
    'GmbH': 'HRB',
    'UG': 'HRB',
    'UG i.G.': 'HRB',
    'AG': 'HRB',
    'e.G.': 'GnR',
    'PartG': 'PR',
}


# possible formats the number can be specified in
_court_re = r'(?P<court>.*)'
_registry_re = r'(?P<registry>%s)' % '|'.join(REGISTRY_TYPES)
_number_re = r'(?P<nr>[1-9][0-9]{0,5})(\s*(?P<x>[A-ZÖ]{1,3}))?'
_formats = [
    _registry_re + r'\s+' + _number_re + r',?\s+' + _court_re + '$',
    _court_re + r',?\s+' + _registry_re + r'\s+' + _number_re + '$',
]


def _split(number):
    """Split the number into a court, registry, register number and
    optionally qualifier."""
    number = clean(number).strip()
    for fmt in _formats:
        m = re.match(fmt, number, flags=re.I | re.U)
        if m:
            return m.group('court').strip(), m.group('registry'), m.group('nr'), m.group('x')
    raise InvalidFormat()


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    court, registry, number, qualifier = _split(number)
    return ' '.join(x for x in [court, registry, number, qualifier] if x)


def validate(number, company_form=None):
    """Check if the number is a valid company registry number. If a
    company_form (eg. GmbH or PartG) is given, the number is validated to
    have the correct registry type."""
    court, registry, number, qualifier = _split(number)
    court = _courts.get(_to_min(court))
    if not court:
        raise InvalidComponent()
    if not isinstance(court, type(number)):  # pragma: no cover (Python 2 code)
        court = court.decode('utf-8')
    if company_form and COMPANY_FORM_REGISTRY_TYPES.get(company_form) != registry:
        raise InvalidComponent()
    return ' '.join(x for x in [court, registry, number, qualifier] if x)


def is_valid(number):
    """Check if the number is a valid company registry number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


# The base URL for performing lookups
_offeneregister_url = 'https://db.offeneregister.de/openregister.json'


def check_offeneregister(number, timeout=30):  # pragma: no cover (not part of normal test suite)
    """Retrieve registration information from the OffeneRegister.de web site.

    This basically returns the JSON response from the web service as a dict.
    It will contain something like the following::

        {
            'companyId': 'U1206_HRB14011',
            'courtCode': 'U1206',
            'courtName': 'Chemnitz',
            'globalId': 'sn_293298',
            'isCurrent': 'True',
            'name': 'Internet hier GmbH Presentation Provider',
            'nativeReferenceNumber': 'Chemnitz HRB 14011',
            'referenceNumberFirstSeen': '2020-06-12',
            'stdRefNo': 'U1206_HRB14011',
            'validFrom': '2020-06-12',
            'validTill': '',
        }

    Will return None if the number is invalid or unknown.
    """
    # this function isn't automatically tested because it would require
    # network access for the tests and unnecessarily load the web service
    import requests
    court, registry, number, qualifier = _split(number)
    response = requests.get(
        _offeneregister_url,
        params={
            'sql': '''
                   select *
                   from ReferenceNumbers
                   join Names on ReferenceNumbers.companyId = Names.companyId
                   where nativeReferenceNumber = :p0
                   limit 1
                   ''',
            'p0': '%s %s %s' % (court, registry, number)},
        timeout=timeout)
    response.raise_for_status()
    try:
        json = response.json()
        return dict(zip(json['columns'], json['rows'][0]))
    except (KeyError, IndexError) as e:  # noqa: F841
        return  # number not found



import logging

from application.model.configuration import (
    EndpointConfig,
)
from typing import Dict

logger = logging.getLogger(__name__)


class GetEndPointsConfigInteractor(object):
    def __init__(
        self,                   # type: GetEndPointsConfigInteractor
        endpoints_config,       # type: Dict[str, EndpointConfig]
    ):
        # type: (...) -> None

        self.endpoints_config = endpoints_config

    def get_loyalty_endpoints_config(
        self,    # type: GetEndPointsConfigInteractor
    ):
        # type: (...) -> Dict[str, bool]

        dict_endpoints_config = {}
        for key, endpoint_config in self.endpoints_config.items():
            dict_endpoints_config[key] = endpoint_config.call_external_endpoint

        return dict_endpoints_config
